using Microsoft.AspNetCore.Mvc;
using Stimulsoft.Report;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Text;
using System.Text.Json;
using System.Text.RegularExpressions;

namespace Stimulosft.Demo.Controllers
{
    public class HomeController : Controller
    {
        private readonly IWebHostEnvironment _hostEnvironment;
        private readonly ILogger<HomeController> _logger;

        public HomeController(ILogger<HomeController> logger, IWebHostEnvironment hostEnvironment)
        {
            _logger = logger;
            _hostEnvironment = hostEnvironment;
        }

        public IActionResult Index()
        {
            return View();
        }

        [HttpGet]
        public async Task<IActionResult> GetReport()
        {
            var Report = new StiReport();
            Report.Load(Path.Combine(_hostEnvironment.WebRootPath + "/report.mrt"));
            return Json(Report.SaveToJsonString());
        }

        [HttpGet]
        public async Task<IActionResult> GetSubReport()
        {
            try
            {
                var report = new StiReport();
                report.Load(Path.Combine(_hostEnvironment.WebRootPath + "/subReport.mrt"));
                return Content(report.SaveToJsonString(), "application/json");
            }
            catch (Exception ex)
            {
                _logger.Log(LogLevel.Error, ex, "Server Error 500");
                return Json(null);
            }
        }
    }

    [Route("/DataAdapters")]
    [ApiController]
    public class StimulsoftController : ControllerBase
    {
        [HttpGet] public async Task GetCommand() => await ProcessRequest();
        [HttpPost] public async Task PostCommand() => await ProcessRequest();

        private JsonSerializerOptions jsonOptions = new JsonSerializerOptions { PropertyNamingPolicy = JsonNamingPolicy.CamelCase };
        private Regex serverCertificateRegex = new Regex(@"Trust\s*Server\s*Certificate\s*=", RegexOptions.Compiled | RegexOptions.IgnoreCase);
        private Regex sslModeRegex = new Regex(@"SSL\s*Mode|SslMode\s*=", RegexOptions.Compiled | RegexOptions.IgnoreCase);

        private async Task ProcessRequest()
        {
            var result = new Result { Success = true };
            var encodeResult = false;

            try
            {
                string inputText;
                using (StreamReader stream = new StreamReader(Request.Body))
                {
                    inputText = await stream.ReadToEndAsync();
                }

                if (!string.IsNullOrEmpty(inputText) && inputText[0] != '{')
                {
                    var buffer = Convert.FromBase64String(ROT13(inputText));
                    inputText = Encoding.UTF8.GetString(buffer);
                    encodeResult = true;
                }

                var command = JsonSerializer.Deserialize<CommandJson>(inputText, jsonOptions);
                if (command.Command == "GetSupportedAdapters")
                {
                    result.Success = true;
                    result.Types = new string[1] { "MS SQL" };
                }
                else
                {
                    switch (command.Database)
                    {
                        case "MS SQL":
                            if (!serverCertificateRegex.IsMatch(command.ConnectionString))
                                command.ConnectionString += (command.ConnectionString.TrimEnd().EndsWith(";") ? "" : ";") + "TrustServerCertificate=true;";
                            result = SQLAdapter.Process(command, new SqlConnection(command.ConnectionString));
                            break;
                        default: result.Success = false; result.Notice = $"Unknown database type [{command.Database}]"; break;
                    }
                }
            }
            catch (Exception e)
            {
                result.Success = false;
                result.Notice = e.Message;
            }

            result.HandlerVersion = "2024.3.3";
            result.CheckVersion = true;

            var contentType = "application/json";
            var resultText = JsonSerializer.Serialize(result, jsonOptions);
            if (encodeResult)
            {
                resultText = ROT13(Convert.ToBase64String(Encoding.UTF8.GetBytes(resultText)));
                contentType = "text/plain";
            }

            Response.Headers.Add("Access-Control-Allow-Origin", "*");
            Response.Headers.Add("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept, Engaged-Auth-Token");
            Response.Headers.Add("Cache-Control", "no-cache");
            Response.ContentType = contentType;
            await Response.WriteAsync(resultText);
        }
        private static string ROT13(string input)
        {
            return string.Join("", input.Select(x => char.IsLetter(x) ? (x >= 65 && x <= 77) || (x >= 97 && x <= 109) ? (char)(x + 13) : (char)(x - 13) : x));
        }
    }

    public class Result
    {
        public bool Success { get; set; }

        public string Notice { get; set; }

        public string[] Columns { get; set; }

        public string[][] Rows { get; set; }

        public string[] Types { get; set; }

        public string AdapterVersion { get; set; }

        public string HandlerVersion { get; set; }

        public bool CheckVersion { get; set; }
    }
    public class CommandJson
    {
        public string Command { get; set; }

        public string ConnectionString { get; set; }

        public string Database { get; set; }

        public string QueryString { get; set; }

        public int Timeout { get; set; }

        public ParameterJson[] Parameters { get; set; }

        public bool EscapeQueryParameters { get; set; }
    }
    public class ParameterJson
    {
        public string Name { get; set; }

        public object Value { get; set; }

        public string TypeGroup { get; set; }

        public string TypeName { get; set; }

        public int NetType { get; set; }

        public int Size { get; set; }
    }
    public class SQLAdapter
    {
        private static DbConnection connection;
        private static DbDataReader reader;
        private static CommandJson command;

        private static Result End(Result result)
        {
            result.AdapterVersion = "2024.3.3";
            try
            {
                if (reader != null) reader.Close();
                if (connection != null) connection.Close();

                return result;
            }
            catch (Exception)
            {
                return result;
            }
        }
        private static Result OnError(string message) => End(new Result { Success = false, Notice = message });
        private static Result Connect()
        {
            try
            {
                connection.Open();
                return OnConnect();
            }
            catch (Exception e)
            {
                return OnError(e.Message);
            }
        }
        private static Result OnConnect()
        {
            if (!String.IsNullOrEmpty(command.QueryString)) return Query();
            else return End(new Result { Success = true });
        }
        private static Result Query()
        {
            try
            {
                var sqlCommand = connection.CreateCommand();
                sqlCommand.CommandType = command.Command == "Execute" ? CommandType.StoredProcedure : CommandType.Text;

                foreach (var parameter in command.Parameters)
                {
                    var sqlParameter = sqlCommand.CreateParameter();
                    sqlParameter.ParameterName = parameter.Name;
                    sqlParameter.DbType = (DbType)parameter.NetType;
                    sqlParameter.Size = parameter.Size;
                    if (sqlParameter.DbType == DbType.Decimal) sqlParameter.Precision = (byte)parameter.Size;
                    sqlParameter.Value = GetValue((JsonElement)parameter.Value, parameter.TypeGroup);
                    sqlCommand.Parameters.Add(sqlParameter);
                }
                reader = sqlCommand.ExecuteReader();
                return OnQuery();
            }
            catch (Exception e)
            {
                return OnError(e.Message);
            }
        }
        private static Result OnQuery()
        {
            var columns = new List<string>();
            var rows = new List<string[]>();
            var types = new List<string>();

            for (var index = 0; index < reader.FieldCount; index++)
            {
                var columnName = reader.GetName(index);
                var columnType = GetType(reader.GetDataTypeName(index));
                if (columnType == "string" && reader.GetFieldType(index).Equals(typeof(byte[])))
                    columnType = "array";

                columns.Add(columnName);
                types.Add(columnType);
            }

            while (reader.Read())
            {
                var row = new string[reader.FieldCount];
                for (var index = 0; index < reader.FieldCount; index++)
                {
                    object value = null;
                    try
                    {
                        if (!reader.IsDBNull(index)) value = reader.GetValue(index);
                    }
                    catch
                    {
                        value = null;
                    }

                    if (value == null) value = "";
                    if (value is DateTime)
                    {
                        row[index] = ((DateTime)value).ToString("yyyy-MM-dd'T'HH:mm:ss.fff");
                        types[index] = "datetime";
                    }
                    else if (value is DateTimeOffset)
                    {
                        row[index] = ((DateTimeOffset)value).ToString("yyyy-MM-dd'T'HH:mm:ss.fffK");
                        types[index] = "datetimeoffset";
                    }
                    else if (value is TimeSpan)
                    {
                        row[index] = Math.Truncate(((TimeSpan)value).TotalHours) + ":" + ((TimeSpan)value).Minutes + ":" + ((TimeSpan)value).Seconds;
                        types[index] = "time";
                    }
                    else
                    {
                        if (types[index] == "array")
                            value = GetBytes(index);

                        row[index] = value.ToString();
                    }
                }
                rows.Add(row);
            }

            return End(new Result { Success = true, Columns = columns.ToArray(), Rows = rows.ToArray(), Types = types.ToArray() });
        }
        private static string GetBytes(int index)
        {
            if (reader.GetFieldValue<object>(index) == DBNull.Value) return "";

            var size = reader.GetBytes(index, 0, null, 0, 0);
            if (size == 0) return "";

            var destination = new MemoryStream();
            var buffer = new byte[8040];
            long offset = 0;
            long read;

            while ((read = reader.GetBytes(index, offset, buffer, 0, buffer.Length)) > 0)
            {
                offset += read;
                destination.Write(buffer, 0, (int)read);
                if (size == offset) break;
            }

            return System.Convert.ToBase64String(destination.ToArray());
        }
        private static string GetType(string dbType)
        {
            if (connection is SqlConnection)
            {
                switch (dbType.ToLowerInvariant())
                {
                    case "uniqueidentifier":
                    case "bigint":
                    case "timestamp":
                    case "int":
                    case "smallint":
                    case "tinyint":
                        return "int";

                    case "decimal":
                    case "money":
                    case "smallmoney":
                    case "float":
                    case "real":
                        return "number";

                    case "datetime":
                    case "date":
                    case "datetime2":
                    case "smalldatetime":
                        return "datetime";

                    case "time":
                        return "time";

                    case "datetimeoffset":
                        return "datetimeoffset";

                    case "bit":
                        return "boolean";

                    case "binary":
                    case "image":
                        return "array";
                }
            }
            return "string";
        }
        private static object GetValue(JsonElement json, string type)
        {
            try
            {
                switch (type)
                {
                    case "string": return json.GetString();
                    case "number": return json.GetDecimal();
                    case "datetime": return json.GetDateTime();
                }
            }
            catch
            {
            }
            return json.GetString();
        }
        public static Result Process(CommandJson command, DbConnection connection)
        {
            SQLAdapter.connection = connection;
            SQLAdapter.command = command;
            return Connect();
        }
    }
}
